import pytest

from picodi import Provide, inject


def test_traceback_should_show_actual_source_code():
    """
    Check that inject decorator don't replace traceback when error occurs in
    decorated function, and still show the actual source code that caused the error.
    """

    def get_resource():
        yield "resource"

    @inject
    def service(dep=Provide(get_resource)):
        assert dep
        1 / 0  # noqa: B018

    with pytest.raises(ZeroDivisionError) as excinfo:
        service()

    assert excinfo.traceback[-1].source.lines[-1].strip() == "1 / 0  # noqa: B018"


async def test_traceback_should_show_actual_source_code_async():
    async def get_resource():
        yield "resource"

    @inject
    async def service(dep=Provide(get_resource)):
        assert dep
        1 / 0  # noqa: B018

    with pytest.raises(ZeroDivisionError) as excinfo:
        await service()

    assert excinfo.traceback[-1].source.lines[-1].strip() == "1 / 0  # noqa: B018"



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest_asyncio

from picodi import registry

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator


@pytest_asyncio.fixture()
async def _picodi_shutdown() -> AsyncGenerator[None]:
    """
    Shutdown dependencies after the test (async version).
    Need for tests consistency.
    """
    yield
    await registry.shutdown()


@pytest_asyncio.fixture()
async def _picodi_init_dependencies(
    picodi_init_dependencies_kwargs: dict | None,
) -> None:
    if picodi_init_dependencies_kwargs is None:
        return
    await registry.init(**picodi_init_dependencies_kwargs)



# type: ignore
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
import sys

sys.path.insert(0, os.path.abspath("../"))

project = "Picodi"
copyright = "2024, yakimka"
author = "yakimka"

with open("../pyproject.toml") as f:
    for line in f:
        if line.startswith("version"):
            version = line.split("=")[1].strip().strip('"')
            break
    else:
        raise ValueError("Version not found in pyproject.toml")

release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.intersphinx",
    "sphinx.ext.doctest",
]
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}
autodoc_member_order = "bysource"
autodoc_inherit_docstrings = False

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_logo = "_static/picodi_logo.png"


signatures_for_replace = {
    "~picodi._scopes.AutoScope": "AutoScope",
    "~picodi._scopes.ContextVarScope": "ContextVarScope",
    "~picodi._scopes.ManualScope": "ManualScope",
    "~picodi._scopes.NullScope": "NullScope",
    "~picodi._scopes.ScopeType": "ScopeType",
    "~picodi._scopes.Scope": "Scope",
    "~picodi._scopes.SingletonScope": "SingletonScope",
    "<class 'picodi._scopes.ManualScope'>": "ManualScope",
    "<class 'picodi._scopes.NullScope'>": "NullScope",
    "<class 'picodi._scopes.SingletonScope'>": "SingletonScope",
    "<class 'picodi._scopes.ContextVarScope'>": "ContextVarScope",
}


def process_signature(app, what, name, obj, options, signature, return_annotation):
    if signature:
        for old, new in signatures_for_replace.items():
            signature = signature.replace(old, new)
    if return_annotation:
        for old, new in signatures_for_replace.items():
            return_annotation = return_annotation.replace(old, new)
    return signature, return_annotation


def setup(app):
    app.connect("autodoc-process-signature", process_signature)



from __future__ import annotations

from collections.abc import Callable, Iterable
from dataclasses import dataclass
from typing import Any, NamedTuple

from picodi._scopes import ManualScope

DependencyCallable = Callable[..., Any]
InitDependencies = (
    Iterable[DependencyCallable] | Callable[[], Iterable[DependencyCallable]]
)
LifespanScopeClass = type[ManualScope] | tuple[type[ManualScope], ...]


class Depends(NamedTuple):
    call: DependencyCallable


@dataclass
class DependNode:
    value: DependencyCallable
    name: str | None
    dependencies: list[DependNode]



from __future__ import annotations

import asyncio
import inspect
import threading
from collections.abc import AsyncGenerator, Awaitable, Callable, Generator
from contextlib import (
    _AsyncGeneratorContextManager,
    _GeneratorContextManager,
    asynccontextmanager,
    contextmanager,
)
from dataclasses import dataclass
from typing import Any, AsyncContextManager, ContextManager, Literal, TypeVar, overload

from picodi._internal import (
    async_injection_context,
    build_depend_tree,
    sync_injection_context,
)
from picodi._scopes import AutoScope, ManualScope, NullScope, ScopeType
from picodi._types import (
    DependencyCallable,
    DependNode,
    Depends,
    InitDependencies,
    LifespanScopeClass,
)
from picodi.support import (
    ExitStack,
    NullAwaitable,
    call_cm_async,
    call_cm_sync,
    is_async_function,
)

T = TypeVar("T")
TC = TypeVar("TC", bound=Callable)


class Storage:
    def __init__(self) -> None:
        self.deps: dict[DependencyCallable, Provider] = {}
        self.overrides: dict[DependencyCallable, DependencyCallable] = {}
        self.touched_dependencies: set[DependencyCallable] = set()
        self.scopes: dict[type[ScopeType], ScopeType] = {}

    def add(
        self,
        dependency: DependencyCallable,
        scope_class: type[ScopeType] = NullScope,
        override: bool = False,
    ) -> None:
        with lock:
            if scope_class not in self.scopes:
                self.scopes[scope_class] = scope_class()

            if dependency not in self.deps or override:
                self.deps[dependency] = Provider.from_dependency(
                    dependency=dependency,
                    scope=self.scopes[scope_class],
                )

    def get(self, dependency: DependencyCallable) -> Provider:
        dependency = self.get_dep_or_override(dependency)
        self.touched_dependencies.add(dependency)
        if dependency not in self.deps:
            self.add(dependency)
        return self.deps[dependency]

    def get_dep_or_override(self, dependency: DependencyCallable) -> DependencyCallable:
        return self.overrides.get(dependency, dependency)

    def get_override(self, dependency: DependencyCallable) -> DependencyCallable | None:
        return self.overrides.get(dependency)

    def get_original(self, override: DependencyCallable) -> DependencyCallable | None:
        for original, overriden in self.overrides.items():
            if overriden == override:
                return original
        return None

    def has_overrides(self) -> bool:
        return bool(self.overrides)

    @overload
    def resolve(
        self,
        dependencies: list[DependencyCallable],
        registry: Registry,
        is_async: Literal[False],
    ) -> ContextManager:
        """
        Return sync context manager that will return tuple of results
        """

    @overload
    def resolve(
        self,
        dependencies: list[DependencyCallable],
        registry: Registry,
        is_async: Literal[True],
    ) -> AsyncContextManager:
        """
        Return async context manager that will return tuple of results
        """

    def resolve(
        self, dependencies: list[DependencyCallable], registry: Registry, is_async: bool
    ) -> AsyncContextManager | ContextManager:
        signature = inspect.Signature(
            parameters=[
                inspect.Parameter(
                    f"dep{i}",
                    inspect.Parameter.POSITIONAL_ONLY,
                    default=Depends(dep),
                )
                for i, dep in enumerate(dependencies, start=1)
            ],
        )
        dependant = DependNode(
            value=lambda *args: args if len(args) > 1 else args[0],
            name=None,
            dependencies=[
                build_depend_tree(dep, name=f"dep{i}", storage=self)
                for i, dep in enumerate(dependencies, start=1)
            ],
        )
        resolver = async_injection_context if is_async else sync_injection_context
        return resolver(
            dependant,
            signature,
            registry,
            args=(),
            kwargs={},
        )


class Registry:
    """
    Manages dependencies and overrides.
    """

    def __init__(self, for_init: InitDependencies | None = None) -> None:
        self._storage = Storage()
        self._for_init: list[InitDependencies] = [for_init] if for_init else []

    def add(
        self,
        dependency: DependencyCallable,
        scope_class: type[ScopeType] = NullScope,
    ) -> None:
        """
        Add a dependency to the registry and set scope_class for it.
        """
        self._storage.add(dependency, scope_class, override=True)

    def add_for_init(self, dependencies: InitDependencies) -> None:
        """
        Add a dependencies to the list of dependencies to initialize.
        """
        if dependencies not in self._for_init:
            self._for_init.append(dependencies)

    def set_scope(
        self, scope_class: type[ScopeType], *, auto_init: bool = False
    ) -> Callable[[TC], TC]:
        """
        Decorator to declare a dependency.
        Should be placed last in the decorator chain (on top).

        :param scope_class: specify the scope class to use it for the dependency.
        :param auto_init: if set to ``True``, the dependency will be added to the list
            of dependencies to initialize. This is useful for dependencies that
            need to be initialized before the application starts.
        """

        def decorator(fn: TC) -> TC:
            self._storage.add(
                fn,
                scope_class=scope_class,
                override=True,
            )
            if auto_init:
                self.add_for_init([fn])
            return fn

        return decorator

    def init(self, dependencies: InitDependencies | None = None) -> Awaitable:
        """
        Call this method to init dependencies. Usually, it should be called
        when your application is starting up.

        This method works both for synchronous and asynchronous dependencies.
        If you call it without ``await``, it will initialize only sync dependencies.
        If you call it ``await init(...)``,
        it will initialize both sync and async dependencies.

        :param dependencies: dependencies to initialize. If this argument
            is passed - init dependencies specified in the registry will be ignored.
        """
        if dependencies is None:
            dependencies = self._for_init_list()
        elif callable(dependencies):
            dependencies = dependencies()

        sync_deps: list[DependencyCallable] = []
        async_deps: list[DependencyCallable] = []
        for dep in dependencies:
            provider = self._storage.get(dep)
            if not isinstance(provider.scope, ManualScope):
                raise ValueError(
                    f"Dependency {dep} is not in ManualScope, "
                    "you cannot initialize it manually."
                )
            if provider.is_async:
                async_deps.append(dep)
            else:
                sync_deps.append(dep)

        if sync_deps:
            call_cm_sync(self._resolve_many(*sync_deps))
        if async_deps:
            return call_cm_async(self._aresolve_many(*async_deps))
        return NullAwaitable()

    def _for_init_list(self) -> list[DependencyCallable]:
        dependencies: list[DependencyCallable] = []
        for item in self._for_init:
            if callable(item):
                item = item()
            dependencies.extend(item)
        return dependencies

    def shutdown(self, scope_class: LifespanScopeClass = ManualScope) -> Awaitable:
        """
        Call this method to close dependencies. Usually, it should be called
        when your application is shut down.

        This method works both for synchronous and asynchronous dependencies.
        If you call it without ``await``, it will shutdown only sync dependencies.
        If you call it ``await shutdown()``, it will shutdown both
        sync and async dependencies.

        If you not pass any arguments,
        it will shutdown subclasses of :class:`ManualScope`.

        :param scope_class: you can specify the scope class to shutdown. If passed -
            only dependencies of this scope class and its subclasses will be shutdown.
        """
        tasks = [
            instance.shutdown(global_key=self.shutdown)  # type: ignore[union-attr]
            for klass, instance in self._storage.scopes.items()
            if issubclass(klass, scope_class)
        ]
        tasks = [task for task in tasks if not isinstance(task, NullAwaitable)]
        return asyncio.gather(*tasks) if tasks else NullAwaitable()

    @contextmanager
    def lifespan(self) -> Generator[None]:
        """
        Context manager to manage the lifespan of the application.
        It will automatically call init and shutdown methods.
        """
        self.init()
        try:
            yield
        finally:
            self.shutdown()

    @asynccontextmanager
    async def alifespan(self) -> AsyncGenerator[None]:
        """
        Async context manager to manage the lifespan of the application.
        It will automatically call init and shutdown methods.
        """
        await self.init()
        try:
            yield
        finally:
            await self.shutdown()

    @property
    def touched(self) -> frozenset[DependencyCallable]:
        """
        Get all dependencies that were used during the picodi lifecycle.
        This method will return a frozenset of dependencies that were resolved.
        It will not include dependencies that were overridden.
        Primarily used for testing purposes.
        For example, you can check that mongo
        database was used in the test and clear it after the test.
        """
        return frozenset(self._storage.touched_dependencies)

    def override(
        self,
        dependency: DependencyCallable,
        new_dependency: DependencyCallable | None,
    ) -> ContextManager[None]:
        """
        Override a dependency with a new one. It can be used as a context manager
        or as a regular method call. New dependency will be
        added to the registry.

        :param dependency: dependency to override
        :param new_dependency: new dependency to use. If explicitly set to ``None``,
            it will remove the override.

        Examples
        --------
        .. code-block:: python

            with registry.override(get_settings, real_settings):
                pass

            registry.override(get_settings, real_settings)
            registry.override(get_settings, None)  # clear override
        """
        if self._storage.get_original(dependency):
            raise ValueError("Cannot override an overridden dependency")

        with lock:
            call_dependency = self._storage.overrides.get(dependency)
            if new_dependency is not None:
                self._storage.add(new_dependency)
                if dependency is new_dependency:
                    raise ValueError("Cannot override a dependency with itself")
                self._storage.overrides[dependency] = new_dependency
            else:
                self._storage.overrides.pop(dependency, None)

        @contextmanager
        def manage_context() -> Generator[None]:
            try:
                yield
            finally:
                self.override(dependency, call_dependency)

        return manage_context()

    @overload
    def resolve(self, dependency: Callable[..., Generator[T]]) -> ContextManager[T]:
        """
        Resolve a dependency that is a generator function synchronously.
        """

    @overload
    def resolve(self, dependency: Callable[..., T]) -> ContextManager[T]:
        """
        Resolve a dependency that is a regular function synchronously.
        """

    def resolve(self, dependency: DependencyCallable) -> ContextManager[Any]:
        """
        Resolve a dependency synchronously. Returns a context manager that will
        return the result of the dependency.
        :param dependency: dependency to resolve.
        :return: sync context manager.
        """
        return self._storage.resolve([dependency], self, is_async=False)

    @overload
    def aresolve(
        self, dependency: Callable[..., Generator[T]]
    ) -> AsyncContextManager[T]:
        """
        Resolve a dependency that is a generator function asynchronously.
        """

    @overload
    def aresolve(
        self, dependency: Callable[..., AsyncGenerator[T]]
    ) -> AsyncContextManager[T]:
        """
        Resolve a dependency that is an async generator function asynchronously.
        """

    @overload
    def aresolve(
        self, dependency: Callable[..., Awaitable[T]]
    ) -> AsyncContextManager[T]:
        """
        Resolve a dependency that is an awaitable asynchronously.
        """

    @overload
    def aresolve(self, dependency: Callable[..., T]) -> AsyncContextManager[T]:
        """
        Resolve a dependency that is a regular function asynchronously.
        """

    def aresolve(self, dependency: DependencyCallable) -> AsyncContextManager[Any]:
        """
        Resolve a dependency asynchronously. Returns a context manager that will
        return the result of the dependency.
        Also can resolve sync dependencies in async context.
        :param dependency: dependency to resolve.
        :return: async context manager.
        """
        return self._storage.resolve([dependency], self, is_async=True)

    def _resolve_many(self, *dependencies: DependencyCallable) -> ContextManager[Any]:
        """
        Internal method to resolve multiple dependencies synchronously.
        """
        return self._storage.resolve(list(dependencies), self, is_async=False)

    def _aresolve_many(
        self, *dependencies: DependencyCallable
    ) -> AsyncContextManager[Any]:
        """
        Internal method to resolve multiple dependencies asynchronously.
        """
        return self._storage.resolve(list(dependencies), self, is_async=True)

    def clear_overrides(self) -> None:
        """
        Clear all overrides. It will remove all overrides, but keep the dependencies.
        """
        self._storage.overrides.clear()

    def clear_touched(self) -> None:
        """
        Clear the touched dependencies.
        It will remove list of all dependencies resolved during the picodi lifecycle.
        """
        self._storage.touched_dependencies.clear()

    def _clear(self) -> None:
        """
        Clear the registry. It will remove all dependencies, overrides
        and touched dependencies.
        This method will not close any dependencies. So you need to manually call
        :func:`shutdown` before this method.
        It is useful for testing purposes, when you want to clear the registry
        and start from scratch.
        """
        self._storage.deps.clear()
        self._storage.overrides.clear()
        self._storage.touched_dependencies.clear()
        self._for_init.clear()


@dataclass(frozen=True)
class Provider:
    dependency: DependencyCallable
    is_async: bool
    scope: ScopeType

    @classmethod
    def from_dependency(
        cls,
        dependency: DependencyCallable,
        scope: ScopeType,
    ) -> Provider:
        return cls(
            dependency=dependency,
            is_async=is_async_function(dependency),
            scope=scope,
        )

    def get_scope(self) -> ScopeType:
        return self.scope

    def resolve_value(
        self,
        exit_stack: ExitStack | None,
        registry: Registry,
        dependant: Callable,
        kwargs: dict[str, Any],
    ) -> Any:
        signature = inspect.signature(self.dependency)
        registry_param = signature.parameters.get("registry")
        if registry_param and registry_param.default is signature.empty:
            kwargs["registry"] = registry

        scope = self.get_scope()
        value_or_gen = self.dependency(**kwargs)
        if self.is_async:

            async def resolve_value_inner() -> Any:
                value_or_gen_ = value_or_gen
                if inspect.iscoroutine(value_or_gen):
                    value_or_gen_ = await value_or_gen_
                if inspect.isasyncgen(value_or_gen_):
                    context_manager = asynccontextmanager(
                        lambda *args, **kwargs: value_or_gen_
                    )
                    if isinstance(scope, AutoScope):
                        assert exit_stack is not None, "exit_stack is required"
                        return await exit_stack.enter_context(context_manager())
                    return await scope.enter(context_manager(), global_key=dependant)
                elif isinstance(value_or_gen_, _AsyncGeneratorContextManager):
                    if isinstance(scope, AutoScope):
                        assert exit_stack is not None, "exit_stack is required"
                        return await exit_stack.enter_context(
                            _recreate_cm(value_or_gen_)
                        )
                    return await scope.enter(
                        _recreate_cm(value_or_gen_), global_key=dependant
                    )
                return value_or_gen_

            return resolve_value_inner()

        if inspect.isgenerator(value_or_gen):
            context_manager = contextmanager(lambda *args, **kwargs: value_or_gen)
            if isinstance(scope, AutoScope):
                assert exit_stack is not None, "exit_stack is required"
                return exit_stack.enter_context(context_manager())
            return scope.enter(context_manager(), global_key=dependant)
        elif isinstance(value_or_gen, _GeneratorContextManager):
            if isinstance(scope, AutoScope):
                assert exit_stack is not None, "exit_stack is required"
                return exit_stack.enter_context(_recreate_cm(value_or_gen))
            return scope.enter(_recreate_cm(value_or_gen), global_key=dependant)
        return value_or_gen


def _recreate_cm(
    gen: AsyncContextManager | ContextManager,
) -> AsyncContextManager | ContextManager:
    return gen._recreate_cm()  # type: ignore[union-attr]  # noqa: SLF001


lock = threading.RLock()
registry = Registry()
registry.__doc__ = """
Picodi registry. You can use it to register dependencies, scopes, overrides,
initialize and shutdown dependencies.
"""



import asyncio
from random import randint

import pytest

from picodi import Provide, SingletonScope, inject, registry


@pytest.mark.repeat(5)
def test_scope_resolving_races(start_race):
    @registry.set_scope(SingletonScope)
    def get_random_int():
        return randint(1, 10000)

    @inject
    def service(num: int = Provide(get_random_int)):
        return num

    results = []

    def actual_test():
        results.append(service())

    start_race(threads_num=8, target=actual_test)

    assert len(set(results)) == 1, results


@pytest.mark.repeat(5)
def test_scope_resolving_races_async(start_race):
    @registry.set_scope(SingletonScope)
    async def get_random_int():
        return randint(1, 10000)

    @inject
    async def service(num: int = Provide(get_random_int)):
        return num

    results = []

    async def main():
        results.append(await service())

    def actual_test():
        asyncio.run(main())

    start_race(threads_num=8, target=actual_test)

    assert len(set(results)) == 1, results



"""
Support module for picodi package.

May be useful for writing your own scopes or other customizations.
"""

from __future__ import annotations

import asyncio
import inspect
from contextlib import AsyncExitStack
from contextlib import ExitStack as SyncExitStack
from typing import TYPE_CHECKING, Any, AsyncContextManager, ContextManager

if TYPE_CHECKING:
    from collections.abc import Awaitable, Generator
    from types import TracebackType


class NullAwaitable:
    """Dummy awaitable that does nothing."""

    def __await__(self) -> Generator[None]:
        yield
        return None


class ExitStack:
    """
    A context manager that combines multiple context managers - both sync and async.

    Under the hood, it uses :class:`python:contextlib.ExitStack`
    for sync context managers and :class:`python:contextlib.AsyncExitStack` for async
    """

    def __init__(self) -> None:
        self._sync_stack = SyncExitStack()
        self._async_stack = AsyncExitStack()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        return self._sync_stack.__exit__(exc_type, exc, traceback)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        res_sync = self._sync_stack.__exit__(exc_type, exc, traceback)
        res_async = await self._async_stack.__aexit__(exc_type, exc, traceback)
        return res_sync and res_async

    def enter_context(self, cm: AsyncContextManager | ContextManager) -> Any:
        """
        Enters a new context manager and adds its ``__[a]exit__()``
        method to the callback stack.
        The return value is the result of the context manager’s
        own ``__[a]enter__()`` method.

        :param cm: context manager to enter.
        :return: Result of the context manager's ``__[a]enter__`` method.
        """
        if isinstance(cm, ContextManager):
            return self._sync_stack.enter_context(cm)
        elif isinstance(cm, AsyncContextManager):
            return self._async_stack.enter_async_context(cm)

        raise TypeError(f"Unsupported context manager: {cm}")  # pragma: no cover

    def close(self, exc: BaseException | None = None) -> Awaitable:
        """
        Immediately unwinds the callback stack,
        invoking callbacks in the reverse order of registration.

        :param exc: exception to be passed to the ``__[a]exit__`` method.
        """
        exc_type = type(exc) if exc is not None else None
        self.__exit__(exc_type, exc, exc.__traceback__ if exc else None)
        if (
            is_async_environment()
            # This is a workaround for the issue for RuntimeWarning
            # "coroutine was never awaited". If we in sync function in async context -
            # don't need to await for async exit if there are no async context managers.
            and self._async_stack._exit_callbacks  # type: ignore # noqa: SLF001
        ):
            return self.__aexit__(exc_type, exc, None)
        return NullAwaitable()


def is_async_environment() -> bool:
    """
    Check if we are in async environment.

    :return: True if we are in async environment, False otherwise.
    """
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return False
    return True


def is_async_function(fn: Any) -> bool:
    """
    Check if the function is async.

    :param fn: function to check.
    :return: True if the function is async, False otherwise.
    """
    if asyncio.iscoroutinefunction(fn) or inspect.isasyncgenfunction(fn):
        return True

    if wrpd := getattr(fn, "__wrapped__", None):
        return is_async_function(wrpd)
    return False


def call_cm_sync(cm: ContextManager) -> Any:
    """
    Call a sync context manager and return its result.

    :param cm: context manager to call.
    :return: Result of the context manager's ``__enter__`` method.
    """
    with cm as res:
        return res


async def call_cm_async(cm: AsyncContextManager) -> Any:
    """
    Call an async context manager and return its result.

    :param cm: context manager to call.
    :return: Result of the context manager's ``__aenter__`` method.
    """
    async with cm as res:
        return res



from __future__ import annotations

import contextlib
import copy
import functools
import inspect
import threading
from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING, Annotated, Any, get_origin

from picodi._scopes import AutoScope, ScopeType
from picodi._types import DependencyCallable, DependNode, Depends
from picodi.support import ExitStack

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Generator

    from picodi._registry import Provider, Registry, Storage


try:
    import fastapi.params
except ImportError:  # pragma: no cover
    fastapi = None  # type: ignore[assignment] # pragma: no cover


StackItem = tuple[DependNode, Iterator[DependNode], dict[str, Any]]


@contextlib.contextmanager
def sync_injection_context(
    dependant: DependNode,
    signature: inspect.Signature,
    registry: Registry,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Generator[Any]:
    gen = wrapper_helper(
        dependant,
        signature,
        is_async=False,
        registry=registry,
        args=args,
        kwargs=kwargs,
    )
    value, action = next(gen)
    while True:
        if action == "result":
            yield value
        try:
            value, action = gen.send(value)
        except StopIteration:
            break
    gen.close()


@contextlib.asynccontextmanager
async def async_injection_context(
    dependant: DependNode,
    signature: inspect.Signature,
    registry: Registry,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> AsyncGenerator[Any]:
    gen = wrapper_helper(
        dependant,
        signature,
        is_async=True,
        registry=registry,
        args=args,
        kwargs=kwargs,
    )
    value, action = next(gen)
    exceptions = []
    while True:
        if inspect.iscoroutine(value):
            try:
                value = await value
            except Exception as e:
                exceptions.append(e)

        # Only yield result if no exceptions occurred
        if action == "result" and not exceptions:
            yield value
        try:
            value, action = gen.send(value)
        except StopIteration:
            break
    if exceptions:
        # TODO use `ExceptionGroup` after dropping 3.10 support
        raise exceptions[0]
    gen.close()


def wrapper_helper(
    dependant: DependNode,
    signature: inspect.Signature,
    is_async: bool,
    registry: Registry,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Generator[Any]:
    exit_stack = ExitStack()
    storage = get_storage_from_registry(registry)
    if _need_patch(dependant, storage):
        dependant = copy.deepcopy(dependant)
        _patch_dependant(dependant, storage)
    bound = signature.bind(*args, **kwargs)
    bound.apply_defaults()
    scopes: set[ScopeType] = set()
    is_root = any(isinstance(value, Depends) for value in bound.arguments.values())

    if is_root:
        stack: list[StackItem] = [(dependant, iter(dependant.dependencies), {})]
        while stack:
            node, deps_iter, dep_kwargs = stack[-1]
            try:
                dep = next(deps_iter)
                stack.append((dep, iter(dep.dependencies), {}))
            except StopIteration:
                if node.name is None:
                    for dep in node.dependencies:
                        assert dep.name is not None
                        if isinstance(bound.arguments.get(dep.name), Depends):
                            bound.arguments[dep.name] = dep_kwargs[dep.name]
                    stack.pop()
                    break

                provider = storage.get(node.value)
                scopes.add(provider.get_scope())
                lazy_call = LazyResolver(
                    provider=provider,
                    kwargs=dep_kwargs,
                    exit_stack=exit_stack,
                    registry=registry,
                    dependant=dependant.value,
                )
                res = yield lazy_call(is_async=is_async), "dependency"
                stack.pop()
                _, _, parent_kwargs = stack[-1]
                parent_kwargs[node.name] = res

    for scope in scopes:
        scope.enter_inject(global_key=dependant.value)

    try:
        result = dependant.value(*bound.args, **bound.kwargs)
    except Exception as e:
        for scope in scopes:
            scope.exit_inject(e, global_key=dependant.value)
            if isinstance(scope, AutoScope):
                yield exit_stack.close(e), "close_scope"
        raise

    if inspect.isgenerator(result):

        @functools.wraps(result)  # type: ignore[arg-type]
        def gen() -> Generator[Any]:
            exception = None
            try:
                yield from result
            except Exception as e:
                exception = e
                raise
            finally:
                for scope in scopes:
                    scope.exit_inject(exception, global_key=dependant.value)
                    if isinstance(scope, AutoScope):
                        exit_stack.close(exception)

        yield gen(), "result"
        return
    elif inspect.isasyncgen(result):

        @functools.wraps(result)  # type: ignore[arg-type]
        async def gen() -> AsyncGenerator[Any]:
            exception = None
            try:
                async for item in result:
                    yield item
            except Exception as e:
                exception = e
                raise
            # TODO use
            #   https://flake8-async.readthedocs.io/en/latest/glossary.html#cancel-scope
            #   https://docs.python.org/3/library/asyncio-task.html#task-cancellation
            finally:
                for scope in scopes:
                    scope.exit_inject(exception, global_key=dependant.value)
                    if isinstance(scope, AutoScope):
                        await exit_stack.close(exception)

        yield gen(), "result"
        return

    yield result, "result"
    for scope in scopes:
        scope.exit_inject(global_key=dependant.value)
        if isinstance(scope, AutoScope):
            yield exit_stack.close(), "close_scope"


def _need_patch(dependant: DependNode, storage: Storage) -> bool:
    if dependant.name is None and not storage.has_overrides():
        return False

    for dep in dependant.dependencies:
        if _need_patch(dep, storage):
            return True

    return bool(dependant.name and storage.get_override(dependant.value))


def _patch_dependant(dependant: DependNode, storage: Storage) -> None:
    for dep in dependant.dependencies:
        _patch_dependant(dep, storage)

    if dependant.name is None:
        return

    if override := storage.get_override(dependant.value):
        dependant.value = override
        override_tree = build_depend_tree(dependant.value, storage=storage)
        dependant.dependencies = override_tree.dependencies


def build_depend_tree(
    dependency: DependencyCallable, *, name: str | None = None, storage: Storage
) -> DependNode:
    signature = inspect.signature(dependency)
    dependencies = []
    for name_, value in signature.parameters.items():
        param_dep = _extract_and_register_dependency_from_parameter(
            value,
            storage,
        )
        if param_dep is not None:
            dependencies.append(
                build_depend_tree(param_dep, name=name_, storage=storage)
            )
    return DependNode(value=dependency, dependencies=dependencies, name=name)


def _extract_and_register_dependency_from_parameter(
    value: inspect.Parameter, storage: Storage
) -> DependencyCallable | None:
    if isinstance(value.default, Depends):
        storage.add(value.default.call)
        return value.default.call

    if fastapi is None:
        return None  # type: ignore[unreachable]  # pragma: no cover
    fastapi_dependency = None
    if isinstance(value.default, fastapi.params.Depends):
        fastapi_dependency = value.default.dependency
    elif get_origin(value.annotation) is Annotated:
        for metadata in value.annotation.__metadata__:
            if isinstance(metadata, fastapi.params.Depends):
                fastapi_dependency = metadata.dependency
                break
    if isinstance(fastapi_dependency, Depends):
        storage.add(fastapi_dependency.call)  # type: ignore[unreachable]
        return fastapi_dependency.call
    return None


class LazyResolver:
    def __init__(
        self,
        provider: Provider,
        kwargs: dict[str, Any] | None = None,
        exit_stack: ExitStack | None = None,
        *,
        registry: Registry,
        dependant: Callable,
    ) -> None:
        self.provider = provider
        self.kwargs = kwargs or {}
        self.exit_stack = exit_stack
        self.registry = registry
        self.dependant = dependant

    def __call__(self, is_async: bool) -> Any:
        call = self._resolve_async if is_async else self._resolve
        return call()

    def _resolve(self) -> Any:
        scope = self.provider.get_scope()
        try:
            value = scope.get(self.provider.dependency, global_key=self.dependant)
        except KeyError:
            if self.provider.is_async:
                value = self.provider.dependency()
            else:
                with lock:
                    try:
                        value = scope.get(
                            self.provider.dependency, global_key=self.dependant
                        )
                    except KeyError:
                        value = self.provider.resolve_value(
                            self.exit_stack,
                            registry=self.registry,
                            dependant=self.dependant,
                            kwargs=self.kwargs,
                        )
                        scope.set(
                            self.provider.dependency, value, global_key=self.dependant
                        )
        return value

    async def _resolve_async(self) -> Any:
        scope = self.provider.get_scope()
        try:
            value = scope.get(self.provider.dependency, global_key=self.dependant)
        except KeyError:
            with lock:
                try:
                    value = scope.get(
                        self.provider.dependency, global_key=self.dependant
                    )
                except KeyError:
                    value = self.provider.resolve_value(
                        self.exit_stack,
                        registry=self.registry,
                        dependant=self.dependant,
                        kwargs=self.kwargs,
                    )
                    if self.provider.is_async:
                        value = await value
                    scope.set(
                        self.provider.dependency, value, global_key=self.dependant
                    )
        return value


lock = threading.RLock()


def get_storage_from_registry(registry: Registry) -> Storage:
    return registry._storage  # noqa: SLF001



from dataclasses import dataclass

import pytest

from picodi import Provide, inject

pytestmark = pytest.mark.benchmark_test


@dataclass(frozen=True)
class MySettings:
    http_client_base_url: str
    database_connection_string: str


my_settings = MySettings(
    http_client_base_url="http://example.com",
    database_connection_string="sqlite:///:memory:",
)


class MyHttpClient:
    def __init__(self, base_url: str):
        self.base_url = base_url


class MyDatabase:
    def __init__(self, connection_string: str):
        self.connection_string = connection_string


class MyRepository:
    def __init__(self, database: MyDatabase):
        self.database = database


class MyService:
    def __init__(self, http_client: MyHttpClient, repository: MyRepository):
        self.http_client = http_client
        self.repository = repository


def test_reference(benchmark):
    def create_http_client() -> MyHttpClient:
        return MyHttpClient(my_settings.http_client_base_url)

    def create_database() -> MyDatabase:
        return MyDatabase(my_settings.database_connection_string)

    def create_repository(database: MyDatabase) -> MyRepository:
        return MyRepository(database)

    def create_service(
        http_client: MyHttpClient, repository: MyRepository
    ) -> MyService:
        return MyService(http_client, repository)

    def create_all():
        http_client = create_http_client()
        database = create_database()
        repository = create_repository(database)
        return create_service(http_client, repository)

    result = benchmark(create_all)

    check_result(result)


def test_picodi_injection(benchmark):
    def get_http_client() -> MyHttpClient:
        return MyHttpClient(my_settings.http_client_base_url)

    def get_database() -> MyDatabase:
        return MyDatabase(my_settings.database_connection_string)

    @inject
    def get_repository(database: MyDatabase = Provide(get_database)) -> MyRepository:
        return MyRepository(database)

    @inject
    def get_service(
        http_client: MyHttpClient = Provide(get_http_client),
        repository: MyRepository = Provide(get_repository),
    ) -> MyService:
        return MyService(http_client, repository)

    @inject
    def create_all(service: MyService = Provide(get_service)) -> MyService:
        return service

    result = benchmark(create_all)

    check_result(result)


def check_result(result):
    __tracebackhide__ = True

    assert isinstance(result, MyService)
    assert result.http_client.base_url == my_settings.http_client_base_url
    connection_string = my_settings.database_connection_string
    assert result.repository.database.connection_string == connection_string



from __future__ import annotations

import threading
from contextvars import ContextVar, Token
from typing import TYPE_CHECKING, Any, AsyncContextManager, ContextManager

import picodi
from picodi.support import ExitStack

if TYPE_CHECKING:
    from collections.abc import Awaitable, Hashable

    from starlette.types import ASGIApp, Receive, Scope, Send


_context: ContextVar[dict[str, Any]] = ContextVar("picodi_starlette_context")
_lock = threading.Lock()


def _get_or_create_context() -> dict[str, Any]:
    try:
        return _context.get()
    except LookupError:
        with _lock:
            # Double check if context was created by another thread
            try:
                return _context.get()
            except LookupError:
                return _new_context()[0]


def _new_context() -> tuple[dict[str, Any], Token]:
    context = {
        "store": {},
        "exit_stack": ExitStack(),
    }
    return context, _context.set(context)


class RequestScope(picodi.ContextVarScope):
    def get(self, key: Hashable, *, global_key: Hashable) -> Any:  # noqa: ARG002
        context = _get_or_create_context()

        try:
            value = context["store"][key]
        except LookupError:
            raise KeyError(key) from None
        return value

    def set(
        self,
        key: Hashable,
        value: Any,
        *,
        global_key: Hashable,  # noqa: ARG002
    ) -> None:
        context = _get_or_create_context()
        context["store"][key] = value

    def enter(
        self,
        context_manager: AsyncContextManager | ContextManager,
        *,
        global_key: Hashable,  # noqa: ARG002
    ) -> Awaitable:
        context = _get_or_create_context()
        exit_stack = context["exit_stack"]
        return exit_stack.enter_context(context_manager)

    def shutdown(
        self, exc: BaseException | None = None, *, global_key: Hashable  # noqa: ARG002
    ) -> Any:
        context = _get_or_create_context()
        context["store"].clear()
        exit_stack = context["exit_stack"]
        return exit_stack.close(exc)


class RequestScopeMiddleware:
    """
    Starlette Pure ASGI Middleware for automatically
    initializing and closing request scoped dependencies
    """

    def __init__(
        self,
        app: ASGIApp,
        *,
        registry: picodi.Registry | None = None,
        dependencies_for_init: picodi.InitDependencies | None = None,
    ) -> None:
        self.app = app
        self._registry = registry or picodi.registry
        self._dependencies_for_init = dependencies_for_init

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        # initialize context in middleware contextvars
        #   this is needed to ensure that context is available and if starlette
        #   will execute view in another thread and copy contextvars - it will
        #   already have our context
        _, token = _new_context()
        if self._dependencies_for_init:
            await self._registry.init(self._dependencies_for_init)
        try:
            await self.app(scope, receive, send)
        finally:
            await self._registry.shutdown(scope_class=RequestScope)
            _context.reset(token)



from __future__ import annotations

from typing import Any

from fastapi import Depends as FastAPIDepends

from picodi import registry
from picodi._types import DependencyCallable, Depends
from picodi.integrations.starlette import RequestScope, RequestScopeMiddleware

__all__ = [
    "Provide",
    "RequestScope",
    "RequestScopeMiddleware",
]


class DependsAsyncCallable(Depends):
    async def __call__(self) -> Any:
        return self


class DependsAsyncStandaloneCallable(Depends):
    async def __call__(self) -> Any:
        async with registry.aresolve(self.call) as result:
            yield result


def Provide(  # noqa: N802
    dependency: DependencyCallable, /, *, wrap: bool = False
) -> Any:
    """
    Drop-in replacement for :func:`picodi.Provide` but for FastAPI.

    :param dependency: callable dependency.
    :param wrap: wrap dependency in ``fastapi.Depends``. In this mode you can use
        ``picodi`` dependencies in FastAPI routes
        without :func:`picodi.inject` decorator.
    """
    if wrap:
        return FastAPIDepends(
            DependsAsyncStandaloneCallable(dependency),  # type: ignore[arg-type]
        )

    return DependsAsyncCallable(dependency)



"""
Helper functions and classes for picodi.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    ContextManager,
    Generic,
    TypeVar,
)

from picodi import Registry
from picodi import registry as default_registry

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Coroutine, Generator

sentinel = object()


class PathNotFoundError(Exception):
    def __init__(self, current_path: str, obj: Any):
        self.current_path = current_path
        self.obj = obj
        super().__init__(f"Can't find path '{current_path}' in {type(obj)} object")


def get_value(path: str, obj: Any, *, default: Any = sentinel) -> Any:
    """
    Get attribute from nested objects.
    Can be useful to avoid passing entire objects like
    settings to functions that only need a small part of it.

    :param path: path to the attribute, separated by dots.
    :param obj: object to search the attribute in.
    :param default: default value to return if the attribute is not found.
        If not provided, an :exc:`PathNotFoundError` is raised.

    :raises PathNotFoundError: if the path is not found in the object
        and default is not provided.

    Example
    -------
    .. code-block:: python

        obj = SimpleNamespace(foo=SimpleNamespace(bar={"baz": 42}))

        get_value("foo.bar.baz", obj)  # Output: 42
        get_value("foo.bar.baz2", obj)  # Output: AttributeError
        get_value("foo.bar.baz2", obj, default=12)  # Output: 12

    """
    if not path:
        raise ValueError("Empty path")
    if not isinstance(path, str):
        raise TypeError("Path must be a string")

    value = obj
    current_path = []
    for part in path.split("."):
        current_path.append(part)
        curr_val = _get_attr(value, part)
        if curr_val is sentinel:
            curr_val = _get_item(value, part)
        if curr_val is sentinel:
            if default is sentinel:
                raise PathNotFoundError(".".join(current_path), obj)
            return default
        value = curr_val

    return value


def _get_attr(obj: Any, attr: str) -> Any:
    try:
        return getattr(obj, attr)
    except AttributeError:
        return sentinel


def _get_item(obj: Any, key: str) -> Any:
    try:
        return obj[key]
    except (KeyError, TypeError):
        return sentinel


T = TypeVar("T")


class resolve(Generic[T]):  # noqa: N801
    """
    Create a context manager from a dependency.

    Can be handy for testing or when you need to create "fabric" dependency
    (dependency that creates other dependencies based on some conditions).

    :param dependency: dependency to create a context manager from.
        Like with :func:`Provide` - don't call the dependency function here,
        just pass it.
    :return: sync or async context manager.

    Example
    -------

    .. code-block:: python

        from picodi.helpers import resolve


        def get_42():
            yield 42


        with resolve(get_42) as val:
            assert val == 42
    """

    def __init__(
        self,
        dependency: Callable[
            [],
            Coroutine[T, None, None] | Generator[T] | AsyncGenerator[T] | T,
        ],
        registry: Registry | None = None,
    ) -> None:
        self.dependency = dependency
        registry = registry or default_registry
        self.sync_cm: ContextManager[T] = registry.resolve(
            dependency,  # type: ignore[arg-type]
        )
        self.async_cm: AsyncContextManager[T] = registry.aresolve(
            dependency,  # type: ignore[arg-type]
        )

    def __enter__(self) -> T:
        return self.sync_cm.__enter__()

    def __exit__(self, *args: Any) -> bool | None:
        return self.sync_cm.__exit__(*args)

    async def __aenter__(self) -> T:
        return await self.async_cm.__aenter__()

    async def __aexit__(self, *args: Any) -> bool | None:
        return await self.async_cm.__aexit__(*args)



from picodi._inject import Provide, inject
from picodi._registry import Registry, registry
from picodi._scopes import (
    AutoScope,
    ContextVarScope,
    ManualScope,
    NullScope,
    Scope,
    ScopeType,
    SingletonScope,
)
from picodi._types import InitDependencies

__all__ = [
    "AutoScope",
    "ContextVarScope",
    "InitDependencies",
    "ManualScope",
    "NullScope",
    "Provide",
    "Registry",
    "Scope",
    "ScopeType",
    "SingletonScope",
    "inject",
    "registry",
]



from __future__ import annotations

import functools
import inspect
from collections.abc import Awaitable, Coroutine
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, cast, overload

from picodi._internal import (
    async_injection_context,
    build_depend_tree,
    get_storage_from_registry,
    sync_injection_context,
)
from picodi._registry import Registry
from picodi._registry import registry as default_registry
from picodi._types import DependencyCallable, Depends

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Generator

T = TypeVar("T")
P = ParamSpec("P")


def Provide(dependency: DependencyCallable, /) -> Any:  # noqa: N802
    """
    Declare a provider.
    It takes a single "dependency" callable (like a function).
    Don't call it directly, picodi will call it for you.

    :param dependency: can be a regular function or a generator with one yield.
        If the dependency is a generator, it will be used as a context manager.
        Any generator that is valid for :func:`python:contextlib.contextmanager`
        can be used as a dependency.

    Example
    -------
    .. code-block:: python

        from picodi import Provide, inject


        def get_db():
            yield "db connection"
            print("closing db connection")


        @inject
        def my_service(db: str = Provide(get_db)):
            assert db == "db connection"
    """
    return Depends(dependency)


@overload
def inject(
    fn: Callable[P, Coroutine[Any, Any, T]],
) -> Callable[P, Coroutine[Any, Any, T]]:
    """Decorator to inject dependencies into an async function."""


@overload
def inject(fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
    """Decorator to inject dependencies into an async function."""


@overload
def inject(fn: Callable[P, AsyncGenerator[T]]) -> Callable[P, AsyncGenerator[T]]:
    """Decorator to inject dependencies into an async generator function."""


@overload
def inject(fn: Callable[P, Generator[T]]) -> Callable[P, Generator[T]]:
    """Decorator to inject dependencies into a generator function."""


@overload
def inject(fn: Callable[P, T]) -> Callable[P, T]:
    """Decorator to inject dependencies into a function."""


@overload
def inject(
    fn: None = None,
    *,
    registry: Registry | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator fabric to inject dependencies into a function."""


def inject(
    fn: Callable[P, T] | None = None, *, registry: Registry | None = None
) -> Callable[P, T] | Callable[[Callable[P, T]], Callable[P, T]]:
    """
    Decorator to inject dependencies into a function.
    Use it in combination with :func:`Provide` to declare dependencies.
    Should be placed first in the decorator chain (on bottom).

    :param fn: function to decorate.

    Example
    -------
    .. code-block:: python

        from picodi import inject, Provide


        @inject
        def my_service(db=Provide(some_dependency_func)):
            pass
    """
    if registry is None:
        registry = default_registry

    def inject_decorator(fn: Callable[P, T]) -> Callable[P, T]:
        signature = inspect.signature(fn)
        storage = get_storage_from_registry(registry)
        dependant = build_depend_tree(fn, storage=storage)

        if inspect.iscoroutinefunction(fn) or inspect.isasyncgenfunction(fn):

            @functools.wraps(fn)
            async def fun_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                async with async_injection_context(
                    dependant,
                    signature,
                    registry,
                    args=args,
                    kwargs=kwargs,
                ) as result:
                    return cast("T", result)

            wrapper = fun_wrapper

            if inspect.isasyncgenfunction(fn):

                @functools.wraps(fn)
                async def gen_wrapper(
                    *args: P.args, **kwargs: P.kwargs
                ) -> AsyncGenerator[T]:
                    result = await fun_wrapper(*args, **kwargs)
                    async for value in result:  # type: ignore[attr-defined]
                        try:
                            yield value
                        except Exception as e:
                            try:
                                await result.athrow(e)  # type: ignore[attr-defined]
                            except StopAsyncIteration:
                                break

                wrapper = gen_wrapper  # type: ignore[assignment]

        else:

            @functools.wraps(fn)
            def fun_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                with sync_injection_context(
                    dependant,
                    signature,
                    registry,
                    args=args,
                    kwargs=kwargs,
                ) as result:
                    return cast("T", result)

            wrapper = fun_wrapper

            if inspect.isgeneratorfunction(fn):

                @functools.wraps(fn)
                def gen_wrapper(*args: P.args, **kwargs: P.kwargs) -> Generator[T]:
                    yield from fun_wrapper(*args, **kwargs)  # type: ignore[misc]

                wrapper = gen_wrapper  # type: ignore[assignment]

        return wrapper  # type: ignore[return-value]

    return inject_decorator if fn is None else inject_decorator(fn)



from __future__ import annotations

from contextlib import ExitStack
from typing import TYPE_CHECKING

import pytest

from picodi import Registry
from picodi import registry as default_registry

if TYPE_CHECKING:
    from collections.abc import Callable, Generator


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line(
        "markers", "picodi_override(dep, override): Picodi deps override"
    )
    config.addinivalue_line(
        "markers",
        "picodi_init_dependencies(dependencies): Helper marker for initializing deps",
    )


@pytest.fixture()
def custom_registries() -> list[Registry]:
    """
    Fixture that returns a list of user-defined registries.
    This is useful for testing multiple registries at once.
    """
    return []


@pytest.fixture()
def registries(
    custom_registries: list[Registry],
) -> list[Registry]:
    """
    Fixture that returns the default registry and any user-defined registries.
    """
    items = list(custom_registries)
    if default_registry not in items:
        items.append(default_registry)
    return items


@pytest.fixture()
def _picodi_shutdown(registries: list[Registry]) -> Generator[None]:
    """
    Shutdown dependencies after the test.
    Need for tests consistency.
    """
    yield
    for registry in registries:
        registry.shutdown()


@pytest.fixture()
def _picodi_clear_touched(registries: list[Registry]) -> Generator[None]:
    """
    Clear touched dependencies after the test.
    Need for tests consistency.
    """
    yield
    for registry in registries:
        registry.clear_touched()


@pytest.fixture()
def picodi_overrides() -> list[tuple[Callable, Callable]]:
    """
    Get overrides from markers that will be used in the test.

    Usually, you don't need to use this fixture directly.
    Use `picodi_override` marker instead. But if you want to stick to
    some custom logic you can inherit from this fixture.
    """
    return []


@pytest.fixture()
def picodi_overrides_from_marks(
    request: pytest.FixtureRequest,
) -> list[tuple[Callable, Callable]]:
    for marker in request.node.iter_markers(name="picodi_override"):
        if len(marker.args) not in (1, 2):
            pytest.fail(
                "picodi_override marker must have 2 arguments: dependency and override "
                "OR 1 argument: iterable of tuples with dependency and override"
            )

        overrides: list[tuple[Callable, Callable]] = (
            marker.args[0] if len(marker.args) == 1 else [marker.args]
        )
        if not isinstance(overrides, (list, tuple)):
            pytest.fail("Overrides must be a list or tuple")
        for dep, override in overrides:
            if not callable(dep) or not callable(override):
                pytest.fail("Dependency and override must be callable")

        return overrides
    return []


@pytest.fixture()
def _picodi_override_setup(
    picodi_overrides: list[tuple[Callable, Callable]],
    picodi_overrides_from_marks: list[tuple[Callable, Callable]],
    registries: list[Registry],
) -> Generator[None]:
    overrides = []
    for item in picodi_overrides_from_marks + picodi_overrides:
        if item not in overrides:
            overrides.append(item)
    with ExitStack() as stack:
        for get_dep, override in overrides:
            for registry in registries:
                stack.enter_context(registry.override(get_dep, override))
        yield


@pytest.fixture()
def picodi_init_dependencies_kwargs(request: pytest.FixtureRequest) -> dict | None:
    for marker in request.node.iter_markers(name="picodi_init_dependencies"):
        if marker.args:
            raise ValueError(
                "init_dependencies marker don't support positional arguments"
            )
        if not marker.kwargs:
            raise ValueError("init_dependencies marker must have keyword arguments")
        return marker.kwargs
    return None


@pytest.fixture()
def _picodi_init_dependencies(
    picodi_init_dependencies_kwargs: dict | None,
    registries: list[Registry],
) -> None:
    if picodi_init_dependencies_kwargs is None:
        return
    for registry in registries:
        registry.init(**picodi_init_dependencies_kwargs)


@pytest.fixture(autouse=True)
def _pytest_autouse_fixture(
    _picodi_override_setup: None,
    _picodi_init_dependencies: None,
    _picodi_clear_touched: None,
    _picodi_shutdown: None,
) -> None:
    """
    Just a fixture to make sure that all the autouse fixtures are used.
    """
